@@ -83,6 +83,9 @@ class CommandConfig:
     #: Whether missing required arguments should be allowed (default: raise an exception when they are missing)
     allow_missing: Bool = False
 
+    #: Whether backtracking is enabled for positionals following params with variable nargs
+    allow_backtrack: Bool = True
+
     # endregion
 
     # region Usage & Help Text Options
@@ -110,6 +113,9 @@ class CommandConfig:
             value = self._show_defaults
         self._show_defaults = ShowDefaults(value)
 
+    #: Whether there should be a visual indicator in help text for the parameters that are members of a given group
+    show_group_tree: Bool = False
+
     # endregion
 
     # region Planned Options
