@@ -19,24 +19,17 @@ try:
 except ImportError:
     from .compat import cached_property
 
-from .config import ShowDefaults
 from .context import ctx
 from .exceptions import ParameterDefinitionError, BadArgument, MissingArgument, InvalidChoice, CommandDefinitionError
-from .exceptions import ParamUsageError, ParamConflict, ParamsMissing, NoActiveContext
-from .formatting import HelpEntryFormatter
+from .exceptions import ParamUsageError, ParamConflict, ParamsMissing, NoActiveContext, UnsupportedAction
+from .formatting.utils import HelpEntryFormatter
 from .nargs import Nargs, NargsValue
-from .utils import (
-    _NotSet,
-    Bool,
-    validate_positional,
-    camel_to_snake_case,
-    get_descriptor_value_type,
-    is_numeric,
-)
+from .utils import _NotSet, Bool, validate_positional, camel_to_snake_case, get_descriptor_value_type, is_numeric
 
 if TYPE_CHECKING:
     from .core import CommandType
     from .commands import Command
+    from .formatting.params import ParamHelpFormatter
 
 __all__ = [
     'Parameter',
@@ -157,18 +150,28 @@ class ParamBase(ABC):
     def __hash__(self) -> int:
         return reduce(xor, map(hash, (self.__class__, self.__name, self.name, self.command)))
 
+    # region Usage / Help Text
+
+    @cached_property
+    def formatter(self) -> 'ParamHelpFormatter':
+        from .formatting.params import ParamHelpFormatter  # Here due to circular dependency
+
+        return ParamHelpFormatter(self)  # noqa
+
     @property
     @abstractmethod
     def show_in_help(self) -> bool:
         raise NotImplementedError
 
-    @abstractmethod
-    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
-        raise NotImplementedError
+    def format_usage(self, *args, **kwargs) -> str:
+        """Convenience method for calling :meth:`.ParamHelpFormatter.format_usage`"""
+        return self.formatter.format_usage(*args, **kwargs)
 
-    @abstractmethod
-    def format_help(self, width: int = 30) -> str:
-        raise NotImplementedError
+    def format_help(self, *args, **kwargs) -> str:
+        """Convenience method for calling :meth:`.ParamHelpFormatter.format_help`"""
+        return self.formatter.format_help(*args, **kwargs)
+
+    # endregion
 
 
 class ParamGroup(ParamBase):
@@ -218,29 +221,7 @@ class ParamGroup(ParamBase):
         self.mutually_exclusive = mutually_exclusive
         self.mutually_dependent = mutually_dependent
 
-    def add(self, param: ParamOrGroup):
-        """Add the given parameter without storing a back-reference.  Primary use case is for help text only groups."""
-        self.members.append(param)
-
-    def register(self, param: ParamOrGroup):
-        if self.mutually_exclusive:
-            if (isinstance(param, BasePositional) and 0 not in param.nargs) or isinstance(param, PassThru):
-                cls_name = param.__class__.__name__
-                raise CommandDefinitionError(
-                    f'Cannot add param={param!r} to {self} - {cls_name} parameters cannot be mutually exclusive'
-                )
-            elif isinstance(param, BaseOption) and param.required:
-                raise CommandDefinitionError(
-                    f'Cannot add param={param!r} to {self} - required parameters cannot be mutually exclusive'
-                    ' (but the group can be required)'
-                )
-
-        self.members.append(param)
-        param.group = self
-
-    def register_all(self, params: Iterable[ParamOrGroup]):
-        for param in params:
-            self.register(param)
+    # region Boilerplate Methods
 
     def __repr__(self) -> str:
         exclusive, dependent = str(self.mutually_exclusive)[0], str(self.mutually_dependent)[0]
@@ -250,13 +231,6 @@ class ParamGroup(ParamBase):
             f' members={members!r}, m.exclusive={exclusive}, m.dependent={dependent}]>'
         )
 
-    @classmethod
-    def active_group(cls) -> Optional['ParamGroup']:
-        try:
-            return cls._local.stack[-1]
-        except (AttributeError, IndexError):
-            return None
-
     def __hash__(self) -> int:
         return super().__hash__()
 
@@ -284,6 +258,19 @@ class ParamGroup(ParamBase):
 
         return self.name < other.name
 
+    def __contains__(self, param: ParamOrGroup) -> bool:
+        """
+        Returns True if the given :class:`Parameter` or :class:`ParamGroup` is a member of this group, False otherwise.
+        """
+        return param in self.members
+
+    def __iter__(self) -> Iterator[ParamOrGroup]:
+        yield from self.members
+
+    # endregion
+
+    # region Active Group Methods
+
     def __enter__(self) -> 'ParamGroup':
         """
         A ParamGroup can be used as a context manager, where all Parameters (and ParamGroups) defined inside the
@@ -300,14 +287,44 @@ class ParamGroup(ParamBase):
         self._local.stack.pop()
         return None
 
-    def __contains__(self, param: ParamOrGroup) -> bool:
-        """
-        Returns True if the given :class:`Parameter` or :class:`ParamGroup` is a member of this group, False otherwise.
-        """
-        return param in self.members
+    @classmethod
+    def active_group(cls) -> Optional['ParamGroup']:
+        try:
+            return cls._local.stack[-1]
+        except (AttributeError, IndexError):
+            return None
 
-    def __iter__(self) -> Iterator[ParamOrGroup]:
-        yield from self.members
+    # endregion
+
+    # region Membership Methods
+
+    def add(self, param: ParamOrGroup):
+        """Add the given parameter without storing a back-reference.  Primary use case is for help text only groups."""
+        self.members.append(param)
+
+    def register(self, param: ParamOrGroup):
+        if self.mutually_exclusive:
+            if (isinstance(param, BasePositional) and 0 not in param.nargs) or isinstance(param, PassThru):
+                cls_name = param.__class__.__name__
+                raise CommandDefinitionError(
+                    f'Cannot add param={param!r} to {self} - {cls_name} parameters cannot be mutually exclusive'
+                )
+            elif isinstance(param, BaseOption) and param.required:
+                raise CommandDefinitionError(
+                    f'Cannot add param={param!r} to {self} - required parameters cannot be mutually exclusive'
+                    ' (but the group can be required)'
+                )
+
+        self.members.append(param)
+        param.group = self
+
+    def register_all(self, params: Iterable[ParamOrGroup]):
+        for param in params:
+            self.register(param)
+
+    # endregion
+
+    # region Argument Handling
 
     def _categorize_params(self) -> Tuple[ParamList, ParamList]:
         """Called after parsing to group this group's members by whether they were provided or not."""
@@ -352,8 +369,13 @@ class ParamGroup(ParamBase):
         if required and not ctx.num_provided(self):
             raise ParamsMissing(missing)
 
+    # endregion
+
+    # region Usage / Help Text
+
     @property
     def contains_positional(self) -> bool:
+        # Used by the help text formatter when grouping parameters / groups
         return any(isinstance(p, BasePositional) for p in self)
 
     @property
@@ -364,48 +386,7 @@ class ParamGroup(ParamBase):
             return self.group.show_in_help
         return True
 
-    def format_description(self, group_type: Bool = True) -> str:
-        description = self.description or f'{self.name} options'
-        if group_type and (self.mutually_exclusive or self.mutually_dependent):
-            description += ' (mutually {})'.format('exclusive' if self.mutually_exclusive else 'dependent')
-        return description
-
-    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
-        choices = ','.join(mem.format_usage(include_meta, full, delim) for mem in self.members)
-        return f'{{{choices}}}'
-
-    def format_help(self, width: int = 30, group_type: Bool = True, clean: Bool = True) -> str:
-        """
-        Prepare the help text for this group.
-
-        :param width: The width of the option/action/command column.
-        :param group_type: Whether the group type should be included in the description if this is a mutually
-          exclusive / dependent group
-        :param clean: If this group only contains other groups or Action or SubCommand parameters, then omit the
-          description.
-        :return: The formatted help text.
-        """
-        parts = [self.format_description(group_type) + ':']
-
-        nested, params = 0, 0
-        for member in self.members:
-            if not member.show_in_help:
-                continue
-
-            if isinstance(member, (ChoiceMap, ParamGroup)):
-                nested += 1
-                parts.append('')  # Add space for readability
-            else:
-                params += 1
-            parts.append(member.format_help(width=width))
-
-        if clean and nested and not params:
-            parts = parts[2:]  # remove description and the first spacer
-
-        if not parts[-1].endswith('\n'):  # ensure a new line separates sections, but avoid extra lines
-            parts.append('')
-
-        return '\n'.join(parts)
+    # endregion
 
 
 class Parameter(ParamBase, ABC):
@@ -416,6 +397,8 @@ class Parameter(ParamBase, ABC):
     otherwise additional handling may be necessary in the parser.
     """
 
+    # region Attributes & Initialization
+
     _actions: FrozenSet[str] = frozenset()
     _positional: bool = False
     _repr_attrs: Optional[Collection[str]] = None
@@ -437,6 +420,7 @@ class Parameter(ParamBase, ABC):
         :param repr_attrs: Additional attributes to include in the repr.
         """
         actions = set(cls._actions)  # Inherit actions from parent
+        actions.update(getattr(cls, '_BasicActionMixin__actions', ()))  # Inherit from mixin, if present
         try:
             actions.update(getattr(cls, f'_{cls.__name__}__actions'))
         except AttributeError:
@@ -492,6 +476,15 @@ class Parameter(ParamBase, ABC):
     def _init_value_factory():
         return _NotSet
 
+    def __set_name__(self, command: 'CommandType', name: str):
+        super().__set_name__(command, name)
+        if self.type is None:
+            annotated_type = get_descriptor_value_type(command, name)
+            if annotated_type is not None:
+                self.type = annotated_type
+
+    # endregion
+
     def __repr__(self) -> str:
         attr_names = ('action', 'const', 'default', 'type', 'choices', 'required', 'hide', 'help')
         extra_attrs = self._repr_attrs
@@ -502,12 +495,7 @@ class Parameter(ParamBase, ABC):
         kwargs = ', '.join(f'{a}={v!r}' for a, v in attrs if v not in (None, _NotSet) and not (a == 'hide' and not v))
         return f'{self.__class__.__name__}({self.name!r}, {kwargs})'
 
-    def __set_name__(self, command: 'CommandType', name: str):
-        super().__set_name__(command, name)
-        if self.type is None:
-            annotated_type = get_descriptor_value_type(command, name)
-            if annotated_type is not None:
-                self.type = annotated_type
+    # region Argument Handling
 
     def __get__(self, command: Optional['Command'], owner: 'CommandType'):
         if command is None:
@@ -630,6 +618,16 @@ class Parameter(ParamBase, ABC):
 
     result = result_value
 
+    def can_pop_counts(self) -> List[int]:  # noqa
+        return []
+
+    def pop_last(self, count: int = 1) -> List[str]:
+        raise UnsupportedAction
+
+    # endregion
+
+    # region Usage / Help Text
+
     @property
     def show_in_help(self) -> bool:
         if self.hide:
@@ -638,38 +636,55 @@ class Parameter(ParamBase, ABC):
             return self.group.show_in_help
         return True
 
-    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
-        return self.usage_metavar
+    # endregion
 
-    def format_help(self, width: int = 30) -> str:
-        usage = self.format_usage(include_meta=True, full=True)
-        description = self.help or ''
-        if _should_add_default(self.default, self.help):
-            pad = ' ' if description else ''
-            description += f'{pad}(default: {self.default})'
-        return HelpEntryFormatter(usage, description, width)()
 
-    @property
-    def usage_metavar(self) -> str:
-        if self.choices:
-            return '{{{}}}'.format(','.join(map(str, self.choices)))
-        elif self.metavar:
-            return self.metavar
-        try:
-            use_type_metavar = ctx.use_type_metavar
-        except NoActiveContext:
-            use_type_metavar = False
-        if use_type_metavar and self.type is not None:
-            t = self.type
-            try:
-                name = t.__name__
-            except AttributeError:
-                pass
-            else:
-                if name != '<lambda>':
-                    return name.upper()
+class BasicActionMixin:
+    action: str
+    nargs: Nargs
 
-        return self.name.upper()
+    @parameter_action
+    def store(self: Parameter, value: Any):
+        ctx.set_parsing_value(self, value)
+
+    @parameter_action
+    def append(self: Parameter, value: Any):
+        ctx.get_parsing_value(self).append(value)
+
+    def _pre_pop_values(self: Parameter):
+        if self.action != 'append' or not self.nargs.variable or self.type not in (None, str):
+            return []
+
+        return ctx.get_parsing_value(self)
+
+    def can_pop_counts(self) -> List[int]:
+        values = self._pre_pop_values()
+        if not values:
+            return []
+
+        n_values = len(values)
+        return [i for i in range(1, n_values) if self.nargs.satisfied(n_values - i)]
+
+    def _reset(self: Union[Parameter, 'BasicActionMixin']) -> List[str]:
+        if self.action != 'append' or self.type not in (None, str):
+            raise UnsupportedAction
+
+        values = ctx.get_parsing_value(self)
+        if not values:
+            return values
+
+        ctx.set_parsing_value(self, self._init_value_factory())
+        ctx._provided[self] = 0
+        return values
+
+    def pop_last(self: Union[Parameter, 'BasicActionMixin'], count: int = 1) -> List[str]:
+        values = self._pre_pop_values()
+        if not values or count >= len(values) or not self.nargs.satisfied(len(values) - count):
+            raise UnsupportedAction
+
+        ctx.set_parsing_value(self, values[:-count])
+        ctx.record_action(self, -count)
+        return values[-count:]
 
 
 # region Positional Parameters
@@ -705,15 +720,8 @@ class BasePositional(Parameter, ABC):
             raise ParameterDefinitionError(f"The 'default' arg is not supported for {cls_name} parameters")
         super().__init__(action, **kwargs)
 
-    def format_basic_usage(self) -> str:
-        return self.format_usage()
 
-    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
-        metavar = self.usage_metavar
-        return metavar if not full or self.nargs == 1 else f'{metavar} [{metavar} ...]'
-
-
-class Positional(BasePositional):
+class Positional(BasicActionMixin, BasePositional):
     """A parameter that must be provided positionally."""
 
     def __init__(
@@ -764,14 +772,6 @@ class Positional(BasePositional):
             if action == 'store':
                 self.default = None if default is _NotSet else default
 
-    @parameter_action
-    def store(self, value: Any):
-        ctx.set_parsing_value(self, value)
-
-    @parameter_action
-    def append(self, value: Any):
-        ctx.get_parsing_value(self).append(value)
-
 
 # endregion
 
@@ -796,9 +796,11 @@ class Choice:
         target_str = f', target={self.target}' if self.choice != self.target else ''
         return f'{self.__class__.__name__}({self.choice!r}{target_str}{help_str})'
 
+    def format_usage(self) -> str:
+        return '(default)' if self.choice is None else self.choice
+
     def format_help(self, width: int = 30, lpad: int = 4) -> str:
-        choice = '(default)' if self.choice is None else self.choice
-        return HelpEntryFormatter(choice, self.help, width, lpad)()
+        return HelpEntryFormatter(self.format_usage(), self.help, width, lpad)()
 
 
 class ChoiceMap(BasePositional):
@@ -852,6 +854,8 @@ class ChoiceMap(BasePositional):
         self.description = description
         self.choices = {}
 
+    # region Choice Registration
+
     def _update_nargs(self):
         try:
             lengths = set(map(len, map(str.split, self.choices)))
@@ -875,6 +879,10 @@ class ChoiceMap(BasePositional):
             prefix = 'Invalid default' if choice is None else f'Invalid choice={choice!r} for'
             raise CommandDefinitionError(f'{prefix} target={target!r} - already assigned to {existing}')
 
+    # endregion
+
+    # region Argument Handling
+
     @parameter_action
     def append(self, value: str):
         values = value.split()
@@ -924,29 +932,15 @@ class ChoiceMap(BasePositional):
         choice = self.result_value()
         return self.choices[choice].target
 
+    # endregion
+
+    # region Usage / Help Text
+
     @property
     def show_in_help(self) -> bool:
         return bool(self.choices)
 
-    @property
-    def usage_metavar(self) -> str:
-        if self.choices:
-            return '{{{}}}'.format(','.join(map(str, filter(None, self.choices))))
-        else:
-            return self.metavar or self.name.upper()
-
-    def format_usage(self, include_meta: Bool = None, full: Bool = None, delim: str = None) -> str:
-        return self.usage_metavar
-
-    def format_help(self, width: int = 30):
-        title = self.title or self._default_title
-        help_entry = HelpEntryFormatter(self.format_usage(), self.description, width, lpad=2)()
-        parts = [f'{title}:', help_entry]
-        for choice in self.choices.values():
-            parts.append(choice.format_help(width, lpad=4))
-
-        parts.append('')
-        return '\n'.join(parts)
+    # endregion
 
 
 class SubCommand(ChoiceMap, title='Subcommands', choice_validation_exc=CommandDefinitionError):
@@ -1163,26 +1157,8 @@ class BaseOption(Parameter, ABC):
     def short_opts(self) -> List[str]:
         return sorted(self._short_opts, key=lambda opt: (-len(opt), opt))
 
-    def format_basic_usage(self) -> str:
-        usage = self.format_usage(True)
-        return usage if self.required else f'[{usage}]'
 
-    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
-        if include_meta:
-            metavar = self.usage_metavar.replace('{', '{{').replace('}', '}}')
-            fmt = '{}' if self.nargs == 0 else f'{{}} [{metavar}]' if 0 in self.nargs else f'{{}} {metavar}'
-            if full:
-                return delim.join(fmt.format(opt) for opt in chain(self.long_opts, self.short_opts))
-            else:
-                return fmt.format(self.long_opts[0])
-        else:
-            if full:
-                return delim.join(chain(self.long_opts, self.short_opts))
-            else:
-                return self.long_opts[0]
-
-
-class Option(BaseOption):
+class Option(BasicActionMixin, BaseOption):
     """A generic option that can be specified as ``--foo bar`` or by using other similar forms."""
 
     def __init__(
@@ -1229,14 +1205,6 @@ class Option(BaseOption):
         if action == 'append':
             self._init_value_factory = list
 
-    @parameter_action
-    def store(self, value: Any):
-        ctx.set_parsing_value(self, value)
-
-    @parameter_action
-    def append(self, value: Any):
-        ctx.get_parsing_value(self).append(value)
-
 
 class Flag(BaseOption, accepts_values=False, accepts_none=True):
     """A (typically boolean) option that does not accept any values."""
@@ -1496,27 +1464,3 @@ class PassThru(Parameter):
     @parameter_action
     def store_all(self, values: Collection[str]):
         ctx.set_parsing_value(self, values)
-
-    def format_basic_usage(self) -> str:
-        usage = self.format_usage()
-        return f'-- {usage}' if self.required else f'[-- {usage}]'
-
-
-# region Helper Functions
-
-
-def _should_add_default(default: Any, help_text: Optional[str]) -> bool:
-    if default is _NotSet:
-        return False
-    sd = ctx.show_defaults
-    if sd.value < 2 or (sd & ShowDefaults.MISSING and help_text and 'default:' in help_text):
-        return False
-    elif sd & ShowDefaults.ANY:
-        return True
-    elif sd & ShowDefaults.NON_EMPTY:
-        return bool(default) or not (default is None or isinstance(default, Collection))
-    else:
-        return bool(default)
-
-
-# endregion
